/*-----------------------------------------------------------------------
 *  Torb: an interface between Tcl and CORBA
 *	
 *  Copyright (C) 1999  Colin Hogben
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Author: Colin Hogben <colin.hogben@bigfoot.com>
 *-----------------------------------------------------------------------*/
/*=======================================================================
 *	$Id: marshal.c,v 1.1 1999/06/07 00:42:58 colin Exp $
 *	Marshal and demarshal values between Tcl and GIOP
 *=======================================================================*/
#include "torb.h"
#include "torb-private.h"

/*-----------------------------------------------------------------------
 *	Convert enum token to its integer (long) value
 *-----------------------------------------------------------------------*/
int torb_enum_value(Tcl_Interp *interp, const char *str, CORBA_TypeCode tc,
		    long *p_long)
{
  int i;
  for (i=0; i<tc->sub_parts; i++) {
    if (strcmp(str, tc->subnames[i]) == 0) {
      *p_long = i;			/* Found it */
      return TCL_OK;
    }
  }
  if (interp) {
    Tcl_AppendResult(interp, "Unknown ", tc->repo_id, " enum value \"",
		     str, "\"", (char*)0);
  }
  return TCL_ERROR;
}

/*-----------------------------------------------------------------------
 *	Test the value of a union case label
 *	
 *-----------------------------------------------------------------------*/
static int find_union_case(Tcl_Interp *interp, CORBA_TypeCode tc, long val,
			   int *p_index)
{
  CORBA_TypeCode dtype = _TORB_TYPECODE_DISCRIMINATOR_TYPE(tc);
  CORBA_TCKind dkind = _TORB_TYPECODE_KIND(dtype);
  int dflt = -1, i;
  switch (dkind) {
  case CORBA_tk_char:
    for (i=0; i<_TORB_TYPECODE_MEMBER_COUNT(tc); i++) {
      CORBA_any *lab = _TORB_TYPECODE_MEMBER_LABEL(tc, i);
      CORBA_TCKind lkind = _TORB_TYPECODE_KIND(lab->_type);
      if (lkind == CORBA_tk_octet) {	/* Signifies default */
	dflt = i;
      } else if (lkind != dkind) {
	ERROR("Enum label type does not match discriminator");
      } else if (val == (long)*(CORBA_char *)lab->_value) {
	*p_index = i;
	return TCL_OK;
      }
    }
    break;
  case CORBA_tk_short:
    for (i=0; i<_TORB_TYPECODE_MEMBER_COUNT(tc); i++) {
      CORBA_any *lab = _TORB_TYPECODE_MEMBER_LABEL(tc, i);
      CORBA_TCKind lkind = _TORB_TYPECODE_KIND(lab->_type);
      if (lkind == CORBA_tk_octet) {	/* Signifies default */
	dflt = i;
      } else if (lkind != dkind) {
	ERROR("Enum label type does not match discriminator");
      } else if (val == (long)*(CORBA_short *)lab->_value) {
	*p_index = i;
	return TCL_OK;
      }
    }
    break;
  case CORBA_tk_ushort:
    for (i=0; i<_TORB_TYPECODE_MEMBER_COUNT(tc); i++) {
      CORBA_any *lab = _TORB_TYPECODE_MEMBER_LABEL(tc, i);
      CORBA_TCKind lkind = _TORB_TYPECODE_KIND(lab->_type);
      if (lkind == CORBA_tk_octet) {	/* Signifies default */
	dflt = i;
      } else if (lkind != dkind) {
	ERROR("Enum label type does not match discriminator");
      } else if (val == (long)*(CORBA_unsigned_short *)lab->_value) {
	*p_index = i;
	return TCL_OK;
      }
    }
    break;
  case CORBA_tk_long:
    for (i=0; i<_TORB_TYPECODE_MEMBER_COUNT(tc); i++) {
      CORBA_any *lab = _TORB_TYPECODE_MEMBER_LABEL(tc, i);
      CORBA_TCKind lkind = _TORB_TYPECODE_KIND(lab->_type);
      if (lkind == CORBA_tk_octet) {	/* Signifies default */
	dflt = i;
      } else if (lkind != dkind) {
	ERROR("Enum label type does not match discriminator");
      } else if (val == (long)*(CORBA_long *)lab->_value) {
	*p_index = i;
	return TCL_OK;
      }
    }
    break;
  case CORBA_tk_ulong:
  case CORBA_tk_enum:
    for (i=0; i<_TORB_TYPECODE_MEMBER_COUNT(tc); i++) {
      CORBA_any *lab = _TORB_TYPECODE_MEMBER_LABEL(tc, i);
      CORBA_TCKind lkind = _TORB_TYPECODE_KIND(lab->_type);
      if (lkind == CORBA_tk_octet) {	/* Signifies default */
	dflt = i;
      } else if (lkind != dkind) {
	ERROR("Enum label type does not match discriminator");
      } else if (val == (long)*(CORBA_unsigned_long *)lab->_value) {
	*p_index = i;
	return TCL_OK;
      }
    }
    break;
  default:
    ERROR("Unknown union label type");
  }
  /*--- No match on value - return default, or -1 if no default */
  *p_index = dflt;
  return TCL_OK;
}

/*-----------------------------------------------------------------------
 *	Use a typecode to walk a Tcl object and marshal it into a GIOP
 *	message buffer.
 *-----------------------------------------------------------------------*/
int torb_marshal_by_typecode(Tcl_Interp *interp, _TORB_PUTBUF_DECL(msg),
			     CORBA_TypeCode tc, Tcl_Obj *obj)
{
  switch (tc->kind) {
  case CORBA_tk_void:
    break;				/* Nothing to do */
  case CORBA_tk_short:
    _TORB_PUTALIGN2(msg);
    TORB_MARSHAL_SHORT(interp, msg, obj, return TCL_ERROR);
    break;
  case CORBA_tk_ushort:
    _TORB_PUTALIGN2(msg);
    TORB_MARSHAL_USHORT(interp, msg, obj, return TCL_ERROR);
    break;
  case CORBA_tk_long:
    _TORB_PUTALIGN4(msg);
    TORB_MARSHAL_LONG(interp, msg, obj, return TCL_ERROR);
    break;
  case CORBA_tk_ulong:
    _TORB_PUTALIGN4(msg);
    TORB_MARSHAL_ULONG(interp, msg, obj, return TCL_ERROR);
    break;
  case CORBA_tk_float:
    _TORB_PUTALIGN4(msg);
    TORB_MARSHAL_FLOAT(interp, msg, obj, return TCL_ERROR);
    break;
  case CORBA_tk_double:
    _TORB_PUTALIGN8(msg);
    TORB_MARSHAL_DOUBLE(interp, msg, obj, return TCL_ERROR);
    break;
  case CORBA_tk_boolean:
    TORB_MARSHAL_BOOLEAN(interp, msg, obj, return TCL_ERROR);
    break;
  case CORBA_tk_char: {
    int len;
    char *str = Tcl_GetStringFromObj(obj, &len);
    if (len != 1) {
      ERROR("char argument must have length 1");
    }
    _TORB_MARSHAL_INT1(msg, (CORBA_char)*str);
  } break;
  case CORBA_tk_octet: {
    /* Octet represented as base-16 value, e.g. "4C" */
    int ival;
    char *str = STR(obj);
    ival = strtoul(str, NULL, 16);	/* FIXME validity check */
    TORB_RANGE(interp, ival,0,255, return TCL_ERROR);
    _TORB_MARSHAL_INT1(msg, ival);
  } break;
  case CORBA_tk_any: {
    int objc;
    Tcl_Obj **objv;
    CORBA_TypeCode anytc;
    CHK(Tcl_ListObjGetElements(interp, obj, &objc, &objv));
    if (objc != 2) {
      ERROR("Any must be length 2");
    }
    CHK(torb_TypeCodeObjGetTc(interp, objv[0], &anytc));
    _TORB_MARSHAL_TYPECODE(msg, anytc);
    CHK(torb_marshal_by_typecode(interp, msg, anytc, objv[1]));
  } break;
  case CORBA_tk_TypeCode:
    TORB_MARSHAL_TYPECODE(interp, msg, obj, return TCL_ERROR);
    break;
  case CORBA_tk_objref:
    TORB_MARSHAL_OBJECT(interp, msg, obj, return TCL_ERROR);
    break;
  case CORBA_tk_string:
    _TORB_PUTALIGN4(msg);
    TORB_MARSHAL_STRING(interp, msg, obj, return TCL_ERROR);
    break;
  case CORBA_tk_longlong:
  case CORBA_tk_ulonglong:
  case CORBA_tk_wchar:
  case CORBA_tk_wstring:
    ERROR("Unimplemented");
    break;
  case CORBA_tk_enum: {
    long eval = -1;
    CHK(torb_enum_value(interp, STR(obj), tc, &eval));
    _TORB_PUTALIGN4(msg);
    _TORB_MARSHAL_INT4(msg, eval);
  } break;
  case CORBA_tk_struct: {
    Tcl_Obj **objv;
    int objc, i;
    CHK(Tcl_ListObjGetElements(interp, obj, &objc, &objv));
    if (objc != tc->sub_parts) {
      ERROR("Wrong length list for struct");
    }
    /* ??? get alignment */
    for (i=0; i<objc; i++) {
      CHK(torb_marshal_by_typecode(interp, msg, tc->subtypes[i], objv[i]));
    }
  } break;
  case CORBA_tk_union: {
    Tcl_Obj **objv;
    int objc;
    CORBA_TypeCode dtype;
    CORBA_TCKind kind;
    long lval;
    int index;
    CHK(Tcl_ListObjGetElements(interp, obj, &objc, &objv));
    if (objc != 2) {
      ERROR("Union value must be list of length 2");
    }
    dtype = _TORB_TYPECODE_DISCRIMINATOR_TYPE(tc);
    kind = _TORB_TYPECODE_KIND(dtype);
    switch (kind) {
    case CORBA_tk_enum:
      CHK(torb_enum_value(interp, STR(objv[0]), dtype, &lval));
      break;
    case CORBA_tk_short:
      TORB_GET_LONG(interp, objv[0], &lval, return TCL_ERROR);
      TORB_RANGE(interp, lval, -32768, 32767, return TCL_ERROR);
      break;
    case CORBA_tk_ushort:
      TORB_GET_LONG(interp, objv[0], &lval, return TCL_ERROR);
      TORB_RANGE(interp, lval, 0, 65535, return TCL_ERROR);
      break;
    case CORBA_tk_long:
    case CORBA_tk_ulong:
      TORB_GET_LONG(interp, objv[0], &lval, return TCL_ERROR);
      break;
    case CORBA_tk_char:
    default:
      ERROR("Oops, I kinda don't know this kind");
    }
    /* ??? get alignment */
    CHK(find_union_case(interp, tc, lval, &index));
    CHK(torb_marshal_by_typecode(interp, msg, dtype, objv[0]));
    if (index >= 0) {
      CORBA_any *lab = _TORB_TYPECODE_MEMBER_LABEL(tc, index);
      CHK(torb_marshal_by_typecode(interp, msg, lab->_type, objv[1]));
    }
  } break;
  default:
    ERROR("Unimplemented typekind to marshal");
  }
  return TCL_OK;
}

/*-----------------------------------------------------------------------
 *	Demarshal
 *-----------------------------------------------------------------------*/
int torb_demarshal_by_typecode(Tcl_Interp *interp, _TORB_GETBUF_DECL(msg),
			       CORBA_TypeCode tc, Tcl_Obj **p_obj)
{
  switch (_TORB_TYPECODE_KIND(tc)) {
  case CORBA_tk_null:
  case CORBA_tk_void:
    /*--- Shouldn't normally get called with tk_void or tk_null?
     * Aha, but you might get one in a CORBA_any! */
    *p_obj = Tcl_NewObj();		/* Empty string */
    break;
  case CORBA_tk_short:
    _TORB_GETALIGN2(msg);
    TORB_DEMARSHAL_SHORT(interp, msg, p_obj, return TCL_ERROR);
    break;
  case CORBA_tk_long:
    _TORB_GETALIGN4(msg);
    TORB_DEMARSHAL_LONG(interp, msg, p_obj, return TCL_ERROR);
    break;
  case CORBA_tk_ushort:
    _TORB_GETALIGN2(msg);
    TORB_DEMARSHAL_USHORT(interp, msg, p_obj, return TCL_ERROR);
    break;
  case CORBA_tk_ulong:
    _TORB_GETALIGN4(msg);
    TORB_DEMARSHAL_ULONG(interp, msg, p_obj, return TCL_ERROR);
    break;
  case CORBA_tk_float:
    _TORB_GETALIGN4(msg);
    TORB_DEMARSHAL_FLOAT(interp, msg, p_obj, return TCL_ERROR);
    break;
  case CORBA_tk_double:
    _TORB_GETALIGN4(msg);
    TORB_DEMARSHAL_DOUBLE(interp, msg, p_obj, return TCL_ERROR);
    break;
  case CORBA_tk_boolean: {
    CORBA_boolean val;
    _TORB_DEMARSHAL_INT1(msg, val);
    *p_obj = Tcl_NewBooleanObj(val);
  } break;
  case CORBA_tk_char: {
    CORBA_char val;
    _TORB_DEMARSHAL_INT1(msg, val);
    *p_obj = Tcl_NewStringObj((char *)&val, 1);
  } break;
  case CORBA_tk_octet: {
    CORBA_octet val;
    char buf[2+1];
    _TORB_DEMARSHAL_INT1(msg, val);
    sprintf(buf, "%02x", (int)val);
    *p_obj = Tcl_NewStringObj(buf, 2);
  } break;
  case CORBA_tk_any: {
    CORBA_TypeCode anytc;
    Tcl_Obj *objv[2];
    _TORB_DEMARSHAL_TYPECODE(msg, anytc, torb_orb);
    if (torb_demarshal_by_typecode(interp, msg, anytc, &objv[1]) != TCL_OK) {
      return TCL_ERROR;
    }
    objv[0] = torb_NewTypeCodeObj(anytc);
    *p_obj = Tcl_NewListObj(2, objv);
  } break;
  case CORBA_tk_TypeCode: {
    CORBA_TypeCode ttc;
    _TORB_DEMARSHAL_TYPECODE(msg, ttc, torb_orb);
    *p_obj = torb_NewTypeCodeObj(ttc);
  } break;
  case CORBA_tk_objref: {
    CORBA_Object obj;
    _TORB_DEMARSHAL_OBJECT(msg, obj, torb_orb);
    *p_obj = torb_NewCorbaObj(interp, obj);
  } break;
  case CORBA_tk_struct: case CORBA_tk_except: {
    int len, i;
    Tcl_Obj **objv;
    len = _TORB_TYPECODE_MEMBER_COUNT(tc);
    objv = ckalloc(len * sizeof(Tcl_Obj *));
    /* ??? get alignment */
    for (i=0; i<len; i++) {
      if (torb_demarshal_by_typecode(interp, msg,
				     _TORB_TYPECODE_MEMBER_TYPE(tc,i),
				     &objv[i]) != TCL_OK) {
	/* ??? free objs got so far */
	ckfree(objv);
	return TCL_ERROR;
      }
    }
    *p_obj = Tcl_NewListObj(len, objv);
  } break;
  case CORBA_tk_union: {
    CORBA_TypeCode dtype = _TORB_TYPECODE_DISCRIMINATOR_TYPE(tc);
    Tcl_Obj *objv[2];
    long lval;
    int index;
    switch (_TORB_TYPECODE_KIND(dtype)) {
    case CORBA_tk_short: {
      CORBA_short c;
      _TORB_GETALIGN2(msg);
      _TORB_DEMARSHAL_INT2(msg, c);
      lval = c;
      objv[0] = Tcl_NewIntObj(c);
    } break;
    case CORBA_tk_ushort: {
      CORBA_unsigned_short c;
      _TORB_GETALIGN2(msg);
      _TORB_DEMARSHAL_INT2(msg, c);
      lval = c;
      objv[0] = Tcl_NewIntObj(c);
    } break;
    case CORBA_tk_long: {
      CORBA_long c;
      _TORB_GETALIGN4(msg);
      _TORB_DEMARSHAL_INT4(msg, c);
      lval = c;
      objv[0] = Tcl_NewLongObj(c);
    } break;
    case CORBA_tk_ulong: {
      CORBA_unsigned_long c;
      _TORB_GETALIGN4(msg);
      _TORB_DEMARSHAL_INT4(msg, c);
      lval = c;
      objv[0] = Tcl_NewLongObj(c);
    } break;
    case CORBA_tk_enum: {
      CORBA_unsigned_long c;
      _TORB_GETALIGN4(msg);
      _TORB_DEMARSHAL_INT4(msg, c);
      lval = c;
      if (c >= _TORB_TYPECODE_MEMBER_COUNT(tc)) {
	ERROR("Enum value outside range");
      }
      objv[0] = Tcl_NewStringObj(_TORB_TYPECODE_MEMBER_NAME(tc, c), -1);
    } break;
    default:
      ERROR("Unknown union discriminator type");
    }
    CHK(find_union_case(interp, tc, lval, &index));
    if (index < 0) {			/* No default so empty body */
      objv[1] = Tcl_NewObj();
    } else {
      CHK(torb_demarshal_by_typecode(interp, msg,
				     _TORB_TYPECODE_MEMBER_TYPE(tc, index),
				     &objv[2]));
    }
    *p_obj = Tcl_NewListObj(2, objv);
  } break;
  case CORBA_tk_enum: {
    CORBA_long val;
    _TORB_GETALIGN4(msg);
    _TORB_DEMARSHAL_INT4(msg,val);
    if (val < 0 || val >= _TORB_TYPECODE_MEMBER_COUNT(tc)) {
      ERROR("Invalid enum off wire");
    }
    *p_obj = Tcl_NewStringObj(_TORB_TYPECODE_MEMBER_NAME(tc,val), -1);
  } break;
  case CORBA_tk_string: {
    CORBA_unsigned_long len;
    _TORB_GETALIGN4(msg);
    _TORB_DEMARSHAL_INT4(msg, len);	/* Includes terminating NUL */
    if (len == 0) {
      *p_obj = Tcl_NewObj();
    } else {
      char *str = ckalloc(len);
      _TORB_DEMARSHAL_STRING(msg,str,len);
      *p_obj = Tcl_NewStringObj(str, len-1);
      ckfree(str);
    }
  } break;
  case CORBA_tk_sequence: {
    CORBA_unsigned_long len, i;
    Tcl_Obj **objv;
    CORBA_TypeCode eltc = _TORB_TYPECODE_MEMBER_TYPE(tc, 0);
    _TORB_GETALIGN4(msg);
    _TORB_DEMARSHAL_INT4(msg,len);
    objv = ckalloc(len * sizeof(Tcl_Obj *));
    for (i=0; i<len; i++) {
      if (torb_demarshal_by_typecode(interp, msg, eltc, &objv[i]) != TCL_OK) {
	/* ??? free objs got so far */
	ckfree(objv);
	return TCL_ERROR;
      }
    }
    *p_obj = Tcl_NewListObj(len, objv);
  } break;
  case CORBA_tk_array: {
    CORBA_unsigned_long len, i;
    Tcl_Obj **objv;
    CORBA_TypeCode eltc = _TORB_TYPECODE_MEMBER_TYPE(tc, 0);
    _TORB_GETALIGN4(msg);
    len = _TORB_TYPECODE_LENGTH(tc);
    objv = ckalloc(len * sizeof(Tcl_Obj *));
    for (i=0; i<len; i++) {
      if (torb_demarshal_by_typecode(interp, msg, eltc, &objv[i]) != TCL_OK) {
	/* ??? free objs got so far */
	ckfree(objv);
	return TCL_ERROR;
      }
    }
    *p_obj = Tcl_NewListObj(len, objv);
  } break;
  case CORBA_tk_alias:
    return torb_demarshal_by_typecode(interp, msg, tc, p_obj);
  case CORBA_tk_Principal:
  default:
    ERROR("Unimplemented kind");
  }
  return TCL_OK;
}

