/*-----------------------------------------------------------------------
 *  Torb: an interface between Tcl and CORBA
 *	
 *  Copyright (C) 1999  Colin Hogben
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Author: Colin Hogben <colin.hogben@bigfoot.com>
 *-----------------------------------------------------------------------*/
/*=======================================================================
 *	$Id: object.c,v 1.1 1999/06/07 00:43:08 colin Exp $
 *	Tcl peer for CORBA object
 *
 *	String representation is the stringified IOR (unimplemented)
 *	_or_ a generated string of the form "torb_obj<n>" for some <n>
 *=======================================================================*/
#include "torb-private.h"
#if ECHO_TEST
#include "echo.h"
#endif
#if HAVE_IR
#include "interface_repository.h"
#endif

static Tcl_FreeInternalRepProc Object_FreeInternalRep;
static Tcl_DupInternalRepProc Object_DupInternalRep;
static Tcl_UpdateStringProc Object_UpdateString;
static Tcl_SetFromAnyProc Object_SetFromAny;
static int object_cmd(ClientData clientData, Tcl_Interp *interp,
		      int objc, Tcl_Obj *CONST objv[]);

/*--- Structure containing objtype methods "is" the object type */
Tcl_ObjType torb_Object_objtype = {
  "CORBA_Object",
  Object_FreeInternalRep,
  Object_DupInternalRep,
  Object_UpdateString,
  Object_SetFromAny,
};

/*-----------------------------------------------------------------------
 *	The Tcl object representing a CORBA object reference consists of
 *	three fields: first is real CORBA_Object, second is the object's
 *	command name, third is the command token.
 *	This is used as the clientData to the object_cmd routine, and
 *	the otherValuePtr of a Tcl_Obj.
 *-----------------------------------------------------------------------*/
typedef struct {
  CORBA_Object object;
  char *cmdName;
  Tcl_Command cmdToken;
} TclCorbaObj;

#define TCO(obj) ((TclCorbaObj *)((obj)->internalRep.otherValuePtr))
#define OBJECT(obj) (TCO(obj)->object)
#define CMDNAME(obj) (TCO(obj)->cmdName)

static int lastCorbaObjNum = 0;

/*-----------------------------------------------------------------------
 *	Build valid internal representation from string form
 *-----------------------------------------------------------------------*/
static int Object_SetFromAny(Tcl_Interp *interp, Tcl_Obj *obj)
{
  char *str, *why;
  int n;
  char c;
  Tcl_CmdInfo cmdInfo;
  
  str = STR(obj);
  DBG1("Object_SetFromAny %s", str);

  if ((sscanf(str, "torb_obj%d%c", &n, &c) != 1 && (why="syntax")) ||
      (! Tcl_GetCommandInfo(interp, str, &cmdInfo) && (why="nocmd")) ||
      (cmdInfo.objProc != object_cmd && (why="objProc"))) {
    Tcl_AppendResult(interp, "Not a CORBA object: ", why, (char*)0);
    return TCL_ERROR;
  }
  /*--- Free old representation's internal form */
  if (obj->typePtr != NULL && obj->typePtr->freeIntRepProc != NULL) {
    obj->typePtr->freeIntRepProc(obj);
  }
  /*--- Now set our type and private data */
  obj->typePtr = &torb_Object_objtype;
  TCO(obj) = cmdInfo.clientData;
  return TCL_OK;
}

/*-----------------------------------------------------------------------
 *	Convert to string form
 *-----------------------------------------------------------------------*/
static void Object_UpdateString(Tcl_Obj *obj)
{
  obj->length = strlen(CMDNAME(obj));
  obj->bytes = ckalloc(obj->length + 1);
  strcpy(obj->bytes, CMDNAME(obj));
}

/*-----------------------------------------------------------------------
 *	Duplicate
 *-----------------------------------------------------------------------*/
static void Object_DupInternalRep(Tcl_Obj *obj, Tcl_Obj *new)
{
  TCO(new) = TCO(obj);
  /* ref count ??? */
}

/*-----------------------------------------------------------------------
 *	Free up memory
 *-----------------------------------------------------------------------*/
static void Object_FreeInternalRep(Tcl_Obj *obj)
{
  /* ref count ??? */
#if 0
  CORBA_Object_release(OBJECT(obj));
  Tcl_DeleteCommandFromToken(interp, TCO(obj)->cmdToken);
  ckfree(TCO(obj)->cmdName);
  ckfree(TCO(obj));
#endif
}

/*=======================================================================
 *	Routines for C code
 *=======================================================================*/
Tcl_Obj *torb_NewCorbaObj(Tcl_Interp *interp, CORBA_Object cobj)
{
  Tcl_Obj *self;
  char name[20];
  int len;
  TclCorbaObj *tco;
  CORBA_Environment ev;

  sprintf(name, "torb_obj%d", ++lastCorbaObjNum);
  DBG1("Tcl_NewCorbaObj %s",name);

  tco = ckalloc(sizeof(TclCorbaObj));
  tco->object = cobj;
  CORBA_Object_duplicate(cobj, &ev);
  len = strlen(name);
  strcpy(tco->cmdName = ckalloc(len + 1), name);
  tco->cmdToken = Tcl_CreateObjCommand(interp, name, object_cmd, tco,
				       NULL/*FIXME deleteProc*/);
  DBG1("created command %s",name);
  self = Tcl_NewObj();
  TCO(self) = tco;
  self->typePtr = &torb_Object_objtype;
  Tcl_InvalidateStringRep(self);
  return self;
}

int torb_CorbaObjGetObject(Tcl_Interp *interp, Tcl_Obj *obj,
			  CORBA_Object *cobjPtr)
{
  if (Tcl_ConvertToType(interp, obj, &torb_Object_objtype) != TCL_OK) {
    return TCL_ERROR;
  }
  *cobjPtr = OBJECT(obj);
  return TCL_OK;
}

/*-----------------------------------------------------------------------
 *	Object command
 *	This routine is shared by all CORBA Object peers.  The client
 *	data is the {CORBA_Object, cmdname, cmdtoken} structure.
 *-----------------------------------------------------------------------*/
static int object_cmd(ClientData clientData, Tcl_Interp *interp,
		      int objc, Tcl_Obj *CONST objv[])
{
  TclCorbaObj *tco = clientData;
  CORBA_Object cobj = tco->object;
  char *opt;

  if (objc < 2) {
    Tcl_AppendResult(interp, "usage: ", STR(objv[0]), " option ?arg?...",
		     (char*)0);
    return TCL_ERROR;
  }
  opt = STR(objv[1]);

  if (strcmp(opt, "_interface")==0) {
    goto unimpl;
  } else if (strcmp(opt, "_is_nil")==0) {
    MUSAGE(0,"");
    goto unimpl;
  } else if (strcmp(opt, "_duplicate")==0) {
    goto unimpl;
  } else if (strcmp(opt, "_release")==0) {
    goto unimpl;
  } else if (strcmp(opt, "_non_existent")==0) {
    MUSAGE(0,"");
    goto unimpl;
  } else if (strcmp(opt, "_is_equivalent")==0) {
    goto unimpl;

  } else if (strcmp(opt, "_hash")==0) {
    CORBA_unsigned_long hash;
    long max;
    MUSAGE(1,"maximum");
    if (Tcl_GetLongFromObj(interp, objv[2], &max) != TCL_OK) return TCL_ERROR;
    EVCHECK(hash = CORBA_Object_hash(cobj, (CORBA_unsigned_long)max, &ev));
    Tcl_SetObjResult(interp, Tcl_NewLongObj((long)hash));

  } else if (strcmp(opt, "get_policy")==0) {
    goto unimpl;
  } else if (strcmp(opt, "get_domain_managers")==0) {
    goto unimpl;

#if ECHO_TEST
  } else if (strcmp(opt, "echoString")==0) {
    /*--- Special to test echo server */
    /*	Echo echoString(in string astring, out long anum); */
    Echo _result;
    CORBA_long anum;
    Tcl_Obj *reslist[2];
    MUSAGE(1,"astring");
    EVCHECK(_result = Echo_echoString(cobj, STR(objv[2]), &anum, &ev));
    reslist[0] = torb_NewCorbaObj(interp, _result);
    reslist[1] = Tcl_NewLongObj(anum);
    Tcl_SetObjResult(interp, Tcl_NewListObj(2, reslist));
    return TCL_OK;
#endif /* ECHO_TEST */

#if HAVE_IR
    /*--- Specials for IR */
  } else if (strcmp(opt, "lookup_id")==0) {
    CORBA_Contained _result;
    Tcl_Obj _reslist[2];
    MUSAGE(1,"id");
    EVCHECK(_result = CORBA_Repository_lookup_id(cobj, STR(objv[2]), &ev));
    Tcl_SetObjResult(interp, Tcl_NewCorbaObj(_result));
  } else if (strcmp(opt, "describe")==0) {
    CORBA_Contained_Description _result;
    MUSAGE(0,"");
    EVCHECK(_result = CORBA_Contained_describe(cobj, &ev));
    _reslist[0] = Tcl_NewIntObj(_result.kind);
    _reslist[1] = any_to_tcl(_result.value);
    Tcl_SetObjResult(interp, Tcl_NewListObj(_reslist, 2));

#endif /* HAVE_IR */

  } else {
    Tcl_AppendResult(interp, "bad option \"", opt, "\": must be ",
		     "_is_nil, _hash", (char*)0);
    return TCL_ERROR;
  }
  return TCL_OK;

 unimpl:
  Tcl_AppendResult(interp, opt, " not yet implemented", (char*)0);
  return TCL_ERROR;
}
