/*-----------------------------------------------------------------------
 *  Torb: an interface between Tcl and CORBA
 *	
 *  Copyright (C) 1999  Colin Hogben
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Author: Colin Hogben <colin.hogben@bigfoot.com>
 *-----------------------------------------------------------------------*/
/*=======================================================================
 *	$Id: torb-orbit.c,v 1.1 1999/06/07 00:43:43 colin Exp $
 *	ORBit-specific back end to torb.
 *	Allows multiple different front-end invocation syntaxes, e.g.
 *	  torb dii -intypes {l s} -outtypes {b} $obj opname $arg1 $arg2
 *	  $obj op $arg1 $arg2
 *	  $obj op -inarg $val1 -ioarg var2 -outarg var3
 *=======================================================================*/
#include "torb.h"

int torb_do_operation_v(Tcl_Interp *interp,
			CORBA_ORB orb,
			CORBA_Object obj,
			const char *opname,
			int incount,
			CORBA_TypeCode *intypes,
			Tcl_Obj *CONST *indata,
			int outcount,
			CORBA_TypeCode *outtypes,
			Tcl_Obj **outdata,
			int excount,
			CORBA_TypeCode *extypes,
			const char *flags,
			va_list ap)
{
  /* Adapted from typical generated stubs. Also see CORBA_Request_send. */
  GIOP_unsigned_long reqid;
  GIOPSendBuffer *sendbuf = NULL;
  GIOPRecvBuffer *recvbuf = NULL;
  struct { CORBA_unsigned_long len; char opname[1]; } *opnameinfo;
  struct iovec opvec;
  GIOPConnection *cnx;
  int i;
  CORBA_Environment ev;

  CORBA_exception_init(&ev);

#if 0
  /* Direct call if implementation in same ORB ... */
  if(obj->servant && obj->vepv && Echo__classid)
    {
      return ((POA_Echo__epv *)obj->vepv[Echo__classid])->echoString(obj->servant, astring, anum, ev);

    }
#endif

  cnx = ORBit_object_get_connection(obj);
 retry_request:
  reqid = giop_get_request_id();
  if((cnx == NULL) || (obj->active_profile == NULL)) {
    CORBA_exception_set_system(&ev, ex_CORBA_COMM_FAILURE, CORBA_COMPLETED_NO);
    goto exc;
  }
  /* variable size operation name so have to allocate dynamically */
  opvec.iov_len = sizeof(CORBA_unsigned_long) + strlen(opname) + 1;
  opvec.iov_base = opnameinfo = g_malloc(opvec.iov_len);
  opnameinfo->len = strlen(opname) + 1;
  strcpy(opnameinfo->opname, opname);

  sendbuf = 
    giop_send_request_buffer_use(cnx,
				 NULL,
				 reqid,
				 CORBA_TRUE,
				 &(obj->active_profile->object_key_vec),
				 &opvec,
				 &ORBit_default_principal_iovec);

  if (!sendbuf) {
    CORBA_exception_set_system(&ev, ex_CORBA_COMM_FAILURE, CORBA_COMPLETED_NO);
    goto exc;
  }


  /* marshal parameters */
  for (i=0; i<incount; i++) {
    if (torb_marshal_by_typecode(interp, sendbuf, intypes[i], indata[i])
	!= TCL_OK) {
      goto fail;
    }
  }

  giop_send_buffer_write(sendbuf);
  giop_send_buffer_unuse(sendbuf);
  recvbuf = giop_recv_reply_buffer_use_2(ORBit_object_get_connection(obj), reqid, TRUE);
  if(recvbuf == NULL || recvbuf->message_buffer.message_header.message_type != GIOP_REPLY) {
    CORBA_exception_set_system(&ev, ex_CORBA_COMM_FAILURE, CORBA_COMPLETED_MAYBE);
    if (recvbuf) giop_recv_buffer_unuse(recvbuf);
    goto exc;
  }
  if(recvbuf->message.u.reply.reply_status != GIOP_NO_EXCEPTION) {
    if(recvbuf->message.u.reply.reply_status == GIOP_LOCATION_FORWARD) {
      if(obj->forward_locations != NULL)
	ORBit_delete_profiles(obj->forward_locations);
      obj->forward_locations = ORBit_demarshal_IOR(recvbuf);
      cnx = ORBit_object_get_forwarded_connection(obj);
      giop_recv_buffer_unuse(recvbuf);

      goto retry_request;
    } else {
      ORBit_handle_exception(recvbuf, &ev, NULL,obj->orb);
      giop_recv_buffer_unuse(recvbuf);
      goto excdone;
    }
  }

  /* demarshal return value(s) */
  for (i=0; i<outcount; i++) {
    if (torb_demarshal_by_typecode(interp, recvbuf, outtypes[i], &outdata[i])
	!= TCL_OK) {
      goto fail;
    }
  }

  giop_recv_buffer_unuse(recvbuf);
  return TCL_OK;

 exc:
 excdone:
 fail:
  return TCL_OK;
}
