/*-----------------------------------------------------------------------
 *  Torb: an interface between Tcl and CORBA
 *	
 *  Copyright (C) 1999  Colin Hogben
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Author: Colin Hogben <colin.hogben@bigfoot.com>
 *-----------------------------------------------------------------------*/
/*=======================================================================
 *	$Id: torb.c,v 1.1 1999/06/07 00:44:01 colin Exp $
 *	Commmand to interface with an ORB
 *=======================================================================*/
#include "torb.h"
#include "torb-private.h"

/*--- Global ORB reference */
CORBA_ORB torb_orb = NULL;

/*-----------------------------------------------------------------------
 *	Convert CORBA error into Tcl error
 *-----------------------------------------------------------------------*/
int torb_econv(Tcl_Interp *interp, CORBA_Environment *ev)
{
  char buf[40];
  if (ev->_major == CORBA_SYSTEM_EXCEPTION) {
    CORBA_SystemException *sys = (CORBA_SystemException *) ev->_params;
    char *val;
    switch (sys->minor) {
    case ex_CORBA_UNKNOWN: val = "UNKNOWN"; break;
    case ex_CORBA_BAD_PARAM: val = "BAD_PARAM"; break;
    case ex_CORBA_NO_MEMORY: val = "NO_MEMORY"; break;
    case ex_CORBA_IMP_LIMIT: val = "IMP_LIMIT"; break;
    case ex_CORBA_COMM_FAILURE: val = "COMM_FAILURE"; break;
    case ex_CORBA_INV_OBJREF: val = "INV_OBJREF"; break;
    case ex_CORBA_NO_PERMISSION: val = "NO_PERMISSION"; break;
    case ex_CORBA_INTERNAL: val = "INTERNAL"; break;
    case ex_CORBA_MARSHAL: val = "MARSHAL"; break;
    case ex_CORBA_INITIALIZE: val = "INITIALIZE"; break;
    case ex_CORBA_NO_IMPLEMENT: val = "NO_IMPLEMENT"; break;
    case ex_CORBA_BAD_TYPECODE: val = "BAD_TYPECODE"; break;
    case ex_CORBA_BAD_OPERATION: val = "BAD_OPERATION"; break;
    case ex_CORBA_NO_RESOURCES: val = "NO_RESOURCES"; break;
    case ex_CORBA_NO_RESPONSE: val = "NO_RESPONSE"; break;
    case ex_CORBA_PERSIST_STORE: val = "PERSIST_STORE"; break;
    case ex_CORBA_BAD_INV_ORDER: val = "BAD_INV_ORDER"; break;
    case ex_CORBA_TRANSIENT: val = "TRANSIENT"; break;
    case ex_CORBA_FREE_MEM: val = "FREE_MEM"; break;
    case ex_CORBA_INV_IDENT: val = "INV_IDENT"; break;
    case ex_CORBA_INV_FLAG: val = "INV_FLAG"; break;
    case ex_CORBA_INTF_REPOS: val = "INTF_REPOS"; break;
    case ex_CORBA_BAD_CONTEXT: val = "BAD_CONTEXT"; break;
    case ex_CORBA_OBJ_ADAPTER: val = "OBJ_ADAPTER"; break;
    case ex_CORBA_DATA_CONVERSION: val = "DATA_CONVERSION"; break;
    case ex_CORBA_OBJECT_NOT_EXIST: val = "OBJECT_NOT_EXIST"; break;
    case ex_CORBA_TRANSACTION_REQUIRED: val = "TRANSACTION_REQUIRED"; break;
    case ex_CORBA_TRANSACTION_ROLLEDBACK: val = "TRANSACTION_ROLLEDBACK"; break;
    case ex_CORBA_INVALID_TRANSACTION: val = "INVALID_TRANSACTION"; break;
    default:
      sprintf(buf, "Unknown_%d", sys->minor);
      val = buf;
    }
    Tcl_AppendResult(interp, "CORBA/SystemException ", val, (char*)0);
  } else if (ev->_major == CORBA_USER_EXCEPTION) {
    Tcl_AppendResult(interp, ev->_repo_id, (char*)0);
  } else {
    sprintf(buf, "%d %#x", ev->_major, ev->_major);
    Tcl_AppendResult(interp, "Unknown CORBA exception type ", buf, (char*)0);
  }
  return TCL_ERROR;
}

/*-----------------------------------------------------------------------
 *	Initialise the ORB
 *-----------------------------------------------------------------------*/
int torb_init(Tcl_Interp *interp, int argc, char **argv)
{
  EVCHECK(torb_orb = CORBA_ORB_init(&argc, argv, "orbit-local-orb", &ev));
  return TCL_OK;
}

/*-----------------------------------------------------------------------
 *	Generic interface to orb-specific back end.
 *	This is the varargs version; the back end is the va_list
 *	version.  We can call that directly too if we want to.
 *	Could have made the backend implement the varargs version, but
 *	Sod's law means we'd need to call it va_list style at
 *	some stage in the future.
 *-----------------------------------------------------------------------*/
int torb_do_operation(Tcl_Interp *interp,
		      CORBA_ORB orb,	/* normally torb_orb */
		      CORBA_Object obj,
		      const char *opname,
		      int incount,
		      CORBA_TypeCode *intypes,
		      Tcl_Obj *CONST *indata,
		      int outcount,
		      CORBA_TypeCode *outtypes,
		      Tcl_Obj **outdata,
		      int excount,
		      CORBA_TypeCode *extypes,
		      const char *flags,
		      ... /*other values dependent on flags*/)
{
  int res;
  va_list ap;
  va_start(ap, flags);
  DBG1("Calling operation %s",opname);
  res = torb_do_operation_v(interp, orb, obj, opname,
			    incount, intypes, indata,
			    outcount, outtypes, outdata,
			    excount, extypes,
			    flags, ap);
  va_end(ap);
  return res;
}


/*-----------------------------------------------------------------------
 *	Now the command itself
 *-----------------------------------------------------------------------*/
static int torb_cmd(ClientData clientData, Tcl_Interp *interp,
		    int objc, Tcl_Obj *CONST objv[])
{
  char *opt;

  if (objc < 2) {
    Tcl_AppendResult(interp, "usage: ", STR(objv[0]), " option ?arg?...",
		     (char*)0);
    return TCL_ERROR;
  }
  opt = STR(objv[1]);

  if (strcmp(opt, "init")==0) {
    char **argv;
    int i;
    if (torb_orb != NULL) {
      Tcl_AppendResult(interp, "ORB already initialised", (char*)NULL);
      return TCL_ERROR;
    }
    argv = (char **) ckalloc(objc * sizeof(char *));
    argv[0] = "torb";
    for (i=0; i<objc-2; i++) {
      argv[i+1] = STR(objv[2+i]);
    }
    argv[objc-1] = NULL;
    return torb_init(interp, objc-1, argv);
    /* only activate POA when needed */

  } else if (strcmp(opt, "dii")==0) {
    CORBA_Object cobj;
    char *opname;
    int incount, outcount, i;
    Tcl_Obj **typev, **results;
    CORBA_TypeCode *intypes, *outtypes;
    
    MUSAGE2(4,-1,"obj opname intypes outtypes ?inval?...");
    INIT();
    CHK(torb_CorbaObjGetObject(interp, objv[2], &cobj));
    opname = STR(objv[3]);
    CHK(Tcl_ListObjGetElements(interp, objv[4], &incount, &typev));
    if (incount == 0) {
      intypes = NULL;
    } else {
      intypes = ckalloc(incount * sizeof(CORBA_TypeCode));
      for (i=0; i<incount; i++) {
	CHK(torb_TypeCodeObjGetTc(interp, typev[i], &intypes[i]));
      }
    }
    CHK(Tcl_ListObjGetElements(interp, objv[5], &outcount, &typev));
    if (outcount == 0) {
      outtypes = NULL;
      results = NULL;
    } else {
      outtypes = ckalloc(outcount * sizeof(CORBA_TypeCode));
      for (i=0; i<outcount; i++) {
	CHK(torb_TypeCodeObjGetTc(interp, typev[i], &outtypes[i]));
      }
      results = ckalloc(outcount * sizeof(Tcl_Obj *));
    }

    if (objc - 6 != incount) {
      ERROR("Mismatsh between argument count and intype list length");
    }
    CHK(torb_do_operation(interp, torb_orb, cobj, opname,
			  incount, intypes, &objv[6],
			  outcount, outtypes, results,
			  0, NULL,	/* FIXME no exceptions */
			  ""));
    if (outcount == 0) {
      ;
    } else if (outcount == 1) {
      /*--- Special case, return single object, not list of length 1 */
      Tcl_SetObjResult(interp, results[0]);
    } else {
      Tcl_SetObjResult(interp, Tcl_NewListObj(outcount, results));
    }

  } else if (strcmp(opt, "string_to_object")==0) {
    CORBA_Object object;

    MUSAGE(1,"ior");
    if (objc != 3) {
      Tcl_AppendResult(interp, "usage: ", STR(objv[0]), " ", opt, " ior",
		       (char*)0);
      return TCL_ERROR;
    }
    INIT();
    EVCHECK(object = CORBA_ORB_string_to_object(torb_orb, STR(objv[2]), &ev));
    /*--- FIXME: may already exist - use hash table? */
    Tcl_SetObjResult(interp, torb_NewCorbaObj(interp, object));

  } else if (strcmp(opt, "object_to_string")==0) {
    char *str;
    CORBA_Object cobj;
    MUSAGE(1,"object");
    INIT();
    CHK(torb_CorbaObjGetObject(interp, objv[2], &cobj));
    EVCHECK(str = CORBA_ORB_object_to_string(torb_orb, cobj, &ev));
    Tcl_SetObjResult(interp, Tcl_NewStringObj(str, -1));
    /* ??? free str */
    
  } else if (strcmp(opt, "resolve_initial_references")==0) {
    CORBA_Object cobj;
    MUSAGE(1,"name");
    INIT();
    EVCHECK(cobj = CORBA_ORB_resolve_initial_references(torb_orb,
							STR(objv[2]), &ev));
    Tcl_SetObjResult(interp, torb_NewCorbaObj(interp, cobj));

  } else if (strcmp(opt, "list_initial_services")==0) {
    CORBA_ORB_ObjectIdList *list;
    int oc, i;
    Tcl_Obj **ov;
    if (objc != 2) {
      Tcl_AppendResult(interp, "usage: ", STR(objv[0]), " ", opt, (char*)0);
      return TCL_ERROR;
    }
    INIT();

    EVCHECK(list = CORBA_ORB_list_initial_services(torb_orb, &ev));

    oc = list->_length;
    ov = ckalloc(oc * sizeof(Tcl_Obj *));
    for (i=0; i<oc; i++) {
      ov[i] = Tcl_NewStringObj(list->_buffer[i], -1);
    }
    Tcl_SetObjResult(interp, Tcl_NewListObj(oc, ov));
    /* FIXME: free CORBA result? */

#if TYPE_TEST
    /*--- Extras for test */
  } else if (strcmp(opt, "type")==0) {
    MUSAGE(1," obj");
    if (objv[2]->typePtr == NULL) {
      Tcl_AppendResult(interp, "(null type)", (char*)0);
    } else if (objv[2]->typePtr->name == NULL) {
      Tcl_AppendResult(interp, "(null name)", (char*)0);
    } else {
      Tcl_AppendResult(interp, objv[2]->typePtr->name, (char*)0);
    }
  } else if (strcmp(opt, "newobj")==0) {
    Tcl_SetObjResult(interp, Tcl_NewObj());
#endif

  } else {
    Tcl_AppendResult(interp, "unknown option \"", opt, "\": must be ",
		     "init, dii, string_to_object, ",
		     "resolve_initial_references, list_initial_services",
		     (char*)0);
    return TCL_ERROR;
  }
  return TCL_OK;
}

/*-----------------------------------------------------------------------
 *	Module initialisation: create the "torb" command
 *-----------------------------------------------------------------------*/
int Torb_Init(Tcl_Interp *interp);
int Torb_Init(Tcl_Interp *interp)
{
  DBG("registering CORBA_TypeCode objtype");
  Tcl_RegisterObjType(&torb_TypeCode_objtype);
  DBG("registering CORBA_Object objtype");
  Tcl_RegisterObjType(&torb_Object_objtype);
  DBG("creating torb command");
  Tcl_CreateObjCommand(interp, "torb", torb_cmd, NULL, NULL/*FIXME torb_delete*/);

  DBG("providing Torb package");
  Tcl_PkgProvide(interp, "Torb", "0.1");
  return TCL_OK;
}
