/*-----------------------------------------------------------------------
 *  Torb: an interface between Tcl and CORBA
 *	
 *  Copyright (C) 1999  Colin Hogben
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Author: Colin Hogben <colin.hogben@bigfoot.com>
 *-----------------------------------------------------------------------*/
/*=======================================================================
 *	$Id: typecode.c,v 1.1 1999/06/07 00:44:23 colin Exp $
 *	Define a new Tcl object type representing a CORBA TypeCode.
 *
 *	String representation is based on that used in TclDii, which is
 *	itself based on Orbix(tm).  Note that I don't have the Orbix
 *	documentation so I can guarantee a certain amount of
 *	incompatibility.  TclIop also uses the same scheme.
 *	Examples:
 *		CORBA_short		s
 *		CORBA_unsigned_long	ul
 *		CORBA_string		0
 *		CORBA_TypeCode		T
 *		struct x {long a,b;}	R~x~a{l},b{l}
 *		enum foo {bar,baz}	E~foo~bar,baz
 *=======================================================================*/
#include "torb-private.h"

static Tcl_FreeInternalRepProc TypeCode_FreeInternalRep;
static Tcl_DupInternalRepProc TypeCode_DupInternalRep;
static Tcl_UpdateStringProc TypeCode_UpdateString;
static Tcl_SetFromAnyProc TypeCode_SetFromAny;
static int TypeCode_from_string(Tcl_Interp *interp, CORBA_TypeCode *p_tc,
				char **p_str);
static void TypeCode_to_dstring(CORBA_TypeCode tc, Tcl_DString *dstr);

/*--- Structure containing objtype methods "is" the object type */
Tcl_ObjType torb_TypeCode_objtype = {
  "CORBA_TypeCode",
  TypeCode_FreeInternalRep,
  TypeCode_DupInternalRep,
  TypeCode_UpdateString,
  TypeCode_SetFromAny,
};

#define TYPECODE(obj) (*(CORBA_TypeCode *)&(obj)->internalRep.otherValuePtr)

/*-----------------------------------------------------------------------
 *	Build valid internal representation from string
 *-----------------------------------------------------------------------*/
static int TypeCode_SetFromAny(Tcl_Interp *interp, Tcl_Obj *obj)
{
  char *str;
  int len;
  CORBA_TypeCode tc;

  /*--- Convert to string */
  str = Tcl_GetStringFromObj(obj, &len);
  if (TypeCode_from_string(interp, &tc, &str) != TCL_OK) {
    return TCL_ERROR;
  }
  /*--- Save real typecode and set object type */
  TYPECODE(obj) = tc;
  obj->typePtr = &torb_TypeCode_objtype;
  return TCL_OK;
}

/*-----------------------------------------------------------------------
 *	Scan a name, delimited by '~', '{', '}' or end of string
 *	Return the length; if non-zero, copy and return the name
 *-----------------------------------------------------------------------*/
static size_t get_name(char *str, char **p_name)
{
  size_t len;
  char *name;

  len = strcspn(str, "~{}");
  if (len == 0) {
    *p_name = NULL;
    return 0;
  }
  name = CORBA_string_alloc(len);
  strncpy(name, str, len);
  name[len] = '\0';
  *p_name = name;
  return len;
}

/*-----------------------------------------------------------------------
 *	Count occurrences of character in typecode string, taking
 *	account of nested typecodes in { }
 *-----------------------------------------------------------------------*/
static int count_chars(const char *str, int wanted)
{
  int parens = 0, count = 0;
  char c;

  while ((c = *str++) != '0') {
    if (c == '{') {
      parens ++;
    } else if (c == '}') {
      if (parens == 0) break;
      parens --;
    } else if (c == wanted && parens == 0) {
      count ++;
    }
  }
  return count;
}

/*-----------------------------------------------------------------------
 *	Create a CORBA_TypeCode from the string (possibly a substring)
 *-----------------------------------------------------------------------*/
static int TypeCode_from_string(Tcl_Interp *interp, CORBA_TypeCode *p_tc,
				char **p_str)
{
  CORBA_TypeCode tc;
  CORBA_Environment ev;
  size_t len;
  char *str = *p_str;

  /* ORBit has TC_CORBA_char etc. but they are const structures */
#define SIMPLETC(t,type) (*(const void **)&(t) = TC_ ## type)
#define LOOKINGAT(c) (*str == (c))
#define EXPECT(c) G_STMT_START{if ((*str++) != (c)) goto bad;}G_STMT_END
#define ATEND() (*str == '\0' || *str == '}')
#define GETNAME(n) G_STMT_START{if ((len = get_name(str, &(n))) == 0) goto bad; str += len;}G_STMT_END
  /* ULONG_MAX/ERANGE is strict ANSI C; some systems e.g. SunOS4 are lax */
#define GETULONG(v) G_STMT_START{unsigned long ul; char *end; if (((ul = strtoul(str, &end, 0)) == ULONG_MAX && errno == ERANGE) || end == str) goto bad; (v) = ul;}G_STMT_END
#define GETLONG(v) G_STMT_START{long l; char *end; if (((((l = strtol(str, &end, 0)) == LONG_MAX) || (l == LONG_MIN)) && errno == ERANGE) || end == str) goto bad; (v) = l;}G_STMT_END
#define GETTYPE(t) G_STMT_START{if (TypeCode_from_string(interp, &(t), &str) != TCL_OK) goto bad;}G_STMT_END  
#define SEQALLOC(p,t,n) G_STMT_START{(p) = t ## Seq__alloc(); (p)->_maximum = (n); (p)->_length = 0; (p)->_buffer = CORBA_sequence_ ## t ## _allocbuf(n); (p)->_release = CORBA_TRUE;}G_STMT_END
#define SEQMEM(p,i) ((p)->_buffer[i])
#define SEQSETLENGTH(p,n) ((p)->_length = (n))

  switch (*str++) {
  case 'n': SIMPLETC(tc,null); break;
  case 'v': SIMPLETC(tc,void); break;
  case 'b': SIMPLETC(tc,CORBA_boolean); break;
  case 'c': SIMPLETC(tc,CORBA_char); break;
  case 'o': SIMPLETC(tc,CORBA_octet); break;
  case 's': SIMPLETC(tc,CORBA_short); break;
  case 'l': SIMPLETC(tc,CORBA_long); break;
#if HAVE_CORBA_LONGLONG
  case 'L': SIMPLETC(tc,CORBA_longlong); break;
#endif
  case 'f': SIMPLETC(tc,CORBA_float); break;
  case 'd': SIMPLETC(tc,CORBA_double); break;
#if HAVE_CORBA_LONGDOUBLE
  case 'D': SIMPLETC(tc,CORBA_longdouble); break;
#endif
  case 'u':
    switch (*str++) {
    case 's': SIMPLETC(tc,CORBA_ushort); break;
    case 'l': SIMPLETC(tc,CORBA_ulong); break;
#if HAVE_CORBA_LONGLONG
    case 'L': SIMPLETC(tc,CORBA_ulonglong); break;
#endif
    default:
      goto bad;
    } break;
  case '0': SIMPLETC(tc,CORBA_string); break;
#if HAVE_CORBA_WCHAR
  case 'w': SIMPLETC(tc,CORBA_wchar); break;
  case 'W': SIMPLETC(tc,CORBA_wstring); break;
#endif
  case 'a': SIMPLETC(tc,CORBA_any); break;
  case 'T': SIMPLETC(tc,CORBA_TypeCode); break;
  case 'P': SIMPLETC(tc,CORBA_Principal); break;
  case 'O':
    if (LOOKINGAT('~')) {
      CORBA_Identifier name;
      GETNAME(name);
    }
    SIMPLETC(tc,CORBA_Object);
    break;
  case 'E':				/* enum */
    {
      CORBA_Identifier name = NULL;
      CORBA_EnumMemberSeq *members = NULL;
      int max, num;

      EXPECT('~');
      GETNAME(name);
      EXPECT('~');
      max = 1 + count_chars(str, ',');
#ifndef CORBA_sequence_CORBA_EnumMember_allocbuf
#define CORBA_sequence_CORBA_EnumMember_allocbuf CORBA_sequence_CORBA_Identifier_allocbuf
#endif
      SEQALLOC(members, CORBA_EnumMember, max);
      for (num = 0; ! ATEND(); ) {
	if (num > 0) EXPECT(',');
	GETNAME(SEQMEM(members, num));
	SEQSETLENGTH(members, ++num);
      }
      INIT();				/* Need orb to be created */
      /*--- FIXME: using name as repo id */
      tc = CORBA_ORB_create_enum_tc(torb_orb, name, name, *members, &ev);
    }
    break;
  case 'S':				/* sequence */
    {
      CORBA_TypeCode elemtc;
      CORBA_unsigned_long bound;

      EXPECT('{');
      GETTYPE(elemtc);
      EXPECT('}');
      EXPECT(',');
      GETULONG(bound);
      INIT();
      tc = CORBA_ORB_create_sequence_tc(torb_orb, bound, elemtc, &ev);
    }
    break;
  case 'A':
    {
      CORBA_TypeCode elemtc;
      CORBA_unsigned_long bound;

      EXPECT('{');
      GETTYPE(elemtc);
      EXPECT('}');
      EXPECT(',');
      GETULONG(bound);
      if (bound == 0) goto bad;		/* Array size must be > 0 */
      INIT();
      tc = CORBA_ORB_create_array_tc(torb_orb, bound, elemtc, &ev);
    }
    break;
  case 'R':				/* struct */
    {
      CORBA_Identifier name;
      CORBA_StructMemberSeq *members = NULL;
      int max, num;

      EXPECT('~');
      GETNAME(name);
      EXPECT('~');
      max = 1 + count_chars(str, ',');
      SEQALLOC(members, CORBA_StructMember, max);
      for (num = 0; ! ATEND(); ) {
	if (num > 0) EXPECT(',');
	GETNAME(SEQMEM(members, num).name);
	EXPECT('{');
	GETTYPE(SEQMEM(members, num).type);
	EXPECT('}');
	/* tyoe_def ??? */
	SEQSETLENGTH(members, ++num);
      }
      INIT();
      /*--- FIXME: should not use name and repo id */
      tc = CORBA_ORB_create_struct_tc(torb_orb, name, name, *members, &ev);
    }
    break;
  case 'X':				/* exception - identical to struct */
    {
      CORBA_Identifier name;
      CORBA_StructMemberSeq *members = NULL;
      int max, num;

      EXPECT('~');
      GETNAME(name);
      EXPECT('~');
      max = 1 + count_chars(str, ',');
      SEQALLOC(members, CORBA_StructMember, max);
      for (num = 0; ! ATEND(); ) {
	if (num > 0) EXPECT(',');
	GETNAME(SEQMEM(members, num).name);
	EXPECT('{');
	GETTYPE(SEQMEM(members, num).type);
	EXPECT('}');
	/* type_def ??? */
	SEQSETLENGTH(members, ++num);
      }
      INIT();
      /*--- FIXME: using name as repo id */
      tc = CORBA_ORB_create_exception_tc(torb_orb, name, name, *members, &ev);
    }
    break;
  case 'U':				/* union */
    {
      CORBA_Identifier name;
      CORBA_TypeCode disc;
      CORBA_UnionMemberSeq *members;
      int max, num;

      EXPECT('~');
      GETNAME(name);
      EXPECT('{');
      GETTYPE(disc);
      EXPECT('}');
      EXPECT('~');
      max = 1 + count_chars(str, ';');
      SEQALLOC(members, CORBA_UnionMember, max);
      for (num = 0; ! ATEND(); ) {
	CORBA_any *label = &SEQMEM(members, num).label;
	if (num > 0) EXPECT(';');
	if (LOOKINGAT('$')) {
	  SIMPLETC(label->_type, octet); /* Is this used by all ORBs? */
	  label->_value = "?";
	  CORBA_any_set_release(label, CORBA_FALSE);
	} else {
	  label->_type = disc;
	  CORBA_Object_duplicate((CORBA_Object)disc, &ev);
	  switch (CORBA_TypeCode_kind(disc, &ev)) {
	  case CORBA_tk_char: {
	    long val;
	    GETLONG(val);
	    /*--- FIXME no check on range */
	    *(CORBA_char *)(label->_value = g_new(CORBA_char, 1)) = val;
	  } break;
	  case CORBA_tk_short: case CORBA_tk_ushort: {
	    long val;
	    GETLONG(val);
	    /*--- FIXME no check on range */
	    *(CORBA_short *)(label->_value = g_new(CORBA_short, 1)) = val;
	  } break;
	  case CORBA_tk_long: {
	    CORBA_long val;
	    GETLONG(val);
	    *(CORBA_long *)(label->_value = g_new(CORBA_long, 1)) = val;
	  } break;
	  case CORBA_tk_ulong: case CORBA_tk_enum: {
	    CORBA_unsigned_long val;
	    GETULONG(val);
	    *(CORBA_long *)(label->_value = g_new(CORBA_long, 1)) = val;
	  } break;
	  default:
	    goto bad;
	  }
	  CORBA_any_set_release(label, CORBA_TRUE);
	}
	EXPECT(' ');
	GETNAME(SEQMEM(members, num).name);
	EXPECT('{');
	GETTYPE(SEQMEM(members, num).type);
	EXPECT('}');
	SEQSETLENGTH(members, ++num);
      }
      INIT();
      /*--- FIXME: using name as repo id */
      tc = CORBA_ORB_create_union_tc(torb_orb, name, name, disc, *members, &ev);
    }
    break;
  default:
    goto bad;
  }
  if (! ATEND()) goto bad;
  CORBA_Object_duplicate((CORBA_Object)tc, &ev);
  *p_tc = tc;
  *p_str = str;
  return TCL_OK;

 bad:
  Tcl_AppendResult(interp, "Invalid TypeCode string", (char*)0);
  return TCL_ERROR;
}

/*-----------------------------------------------------------------------
 *	Convert to string form.
 *	Not needed for user-supplied typecodes, since we never call
 *	Tcl_InvalidateStringRep.  But we can get typecodes returned
 *	from elsewhere.
 *-----------------------------------------------------------------------*/
static void TypeCode_UpdateString(Tcl_Obj *obj)
{
  Tcl_DString dstr;

  Tcl_DStringInit(&dstr);
  /*--- Do the (possibly recursive) conversion into a DString */
  TypeCode_to_dstring(TYPECODE(obj), &dstr);
  /*--- More efficient but less clean would be to extract the
   * string out of the DString directly if has been malloc'ed */
  obj->length = Tcl_DStringLength(&dstr);
  obj->bytes = Tcl_Alloc(obj->length + 1);
  strcpy(obj->bytes, Tcl_DStringValue(&dstr));
  Tcl_DStringFree(&dstr);
}

static void TypeCode_to_dstring(CORBA_TypeCode tc, Tcl_DString *dstr)
{
  unsigned long i;
  char buf[24];

#define PUT(s) Tcl_DStringAppend(dstr, s, -1)
#define PUTTYPE(t) TypeCode_to_dstring(t, dstr)

  switch (tc->kind) {
  case CORBA_tk_null: PUT("v"); break;
  case CORBA_tk_void: PUT("v"); break;
  case CORBA_tk_short: PUT("s"); break;
  case CORBA_tk_long: PUT("l"); break;
  case CORBA_tk_ushort: PUT("us"); break;
  case CORBA_tk_ulong: PUT("ul"); break;
  case CORBA_tk_float: PUT("f"); break;
  case CORBA_tk_double: PUT("d"); break;
  case CORBA_tk_boolean: PUT("b"); break;
  case CORBA_tk_char: PUT("c"); break;
  case CORBA_tk_octet: PUT("o"); break;
  case CORBA_tk_any: PUT("a"); break;
  case CORBA_tk_TypeCode: PUT("T"); break;
  case CORBA_tk_Principal: PUT("P"); break;
  case CORBA_tk_objref: PUT("O"); break;
  case CORBA_tk_struct:
    PUT("R~");
  structmembers:			/* common to struct and exception... */
    PUT(tc->name);			/* or id? */
    for (i=0; i<tc->sub_parts; i++) {
      PUT(i==0 ? "~" : ",");
      PUT(tc->subnames[i]);
      PUT("{");
      PUTTYPE(tc->subtypes[i]);
      PUT("}");
    }
    break;
  case CORBA_tk_union:
    PUT("U~");
    PUT(tc->name);			/* or id? */
    PUT("{");
    PUTTYPE(tc->discriminator);
    PUT("}");
    for (i=0; i<tc->sub_parts; i++) {
      CORBA_any *lab = &tc->sublabels[i];
      PUT(i==0 ? "~" : ";");
      switch (lab->_type->kind) {
      case CORBA_tk_octet:		/* signifies default */
	strcpy(buf, "$");
	break;
      case CORBA_tk_char:
	sprintf(buf, "%d", (int)*(CORBA_char *)lab->_value);
	break;
      case CORBA_tk_short:
	sprintf(buf, "%hd", *(CORBA_short *)lab->_value);
	break;
      case CORBA_tk_ushort:
	sprintf(buf, "%hu", *(CORBA_unsigned_short *)lab->_value);
	break;
      case CORBA_tk_long:
	sprintf(buf, "%ld", (long)*(CORBA_long *)lab->_value);
	break;
      case CORBA_tk_ulong:
      case CORBA_tk_enum:
	sprintf(buf, "%lu", (unsigned long)*(CORBA_unsigned_long *)lab->_value);
	break;
      default:
	PUT("?");			/* Shouldn't happen */
      }
      PUT(buf);
      PUT(" ");
      PUT(tc->subnames[i]);
      PUT("{");
      PUTTYPE(tc->subtypes[i]);
      PUT("}");
    }
    break;
  case CORBA_tk_enum:
    PUT("E~");
    PUT(tc->name);
    for (i=0; i<tc->sub_parts; i++) {
      PUT(i==0 ? "~" : ",");
      PUT(tc->subnames[i]);
    }
    break;
  case CORBA_tk_string: PUT("0"); break;
  case CORBA_tk_sequence:
    PUT("S{");
    PUTTYPE(tc->subtypes[0]);
    PUT("},");
    sprintf(buf, "%lu", (unsigned long)tc->length);
    break;
  case CORBA_tk_array:
    PUT("A{");
    PUTTYPE(tc->subtypes[0]);
    PUT("},");
    sprintf(buf, "%lu", (unsigned long)tc->length);
    break;
  case CORBA_tk_alias:
    PUTTYPE(tc->subtypes[0]);
    break;
  case CORBA_tk_except:
    PUT("X~");
    goto structmembers;
    /*--- These are my invention */
#if HAVE_CORBA_LONGLONG
  case CORBA_tk_longlong: PUT("L"); break;
  case CORBA_tk_ulonglong: PUT("uL"); break;
#endif
#if HAVE_CORBA_LONGDOUBLE
  case CORBA_tk_longdouble: PUT("D"); break;
#endif
#if HAVE_CORBA_WCHAR
  case CORBA_tk_wchar: PUT("w"); break;
  case CORBA_tk_wstring: PUT("W"); break;
#endif
#if HAVE_CORBA_FIXED
  case CORBA_tk_fixed: PUT("f"); break;
#endif
  default:
    /* CORBA_tk_recursive Should only occur when (de)marshalling */
    /* CORBA_tk_last is only a sentinel (ORBit) */
    /*--- Should fault unknown values, but Tcl doesn't want to know */
    PUT("?");
  }
#undef PUT
#undef PUTTYPE
}

/*-----------------------------------------------------------------------
 *	Duplicate, i.e. create another TCL object equal to the first
 *-----------------------------------------------------------------------*/
static void TypeCode_DupInternalRep(Tcl_Obj *obj, Tcl_Obj *new)
{
  CORBA_Environment ev;
  TYPECODE(new) = TYPECODE(obj);
  CORBA_Object_duplicate((CORBA_Object)TYPECODE(new), &ev);
}

/*-----------------------------------------------------------------------
 *	Free up memory associated with it
 *-----------------------------------------------------------------------*/
static void TypeCode_FreeInternalRep(Tcl_Obj *obj)
{
  CORBA_Environment ev;
  CORBA_Object_release((CORBA_Object)TYPECODE(obj), &ev);
}

/*=======================================================================
 *	Routines available to external C code
 *=======================================================================*/
Tcl_Obj *torb_NewTypeCodeObj(CORBA_TypeCode tc)
{
  Tcl_Obj *self;
  CORBA_Environment ev;
  self = Tcl_NewObj();
  TYPECODE(self) = tc;
  CORBA_Object_duplicate((CORBA_Object)tc, &ev);
  Tcl_InvalidateStringRep(self);
  return self;
}

int torb_TypeCodeObjGetTc(Tcl_Interp *interp, Tcl_Obj *obj,
			  CORBA_TypeCode *p_tc)
{
  if (Tcl_ConvertToType(interp, obj, &torb_TypeCode_objtype) != TCL_OK) {
    return TCL_ERROR;
  }
  *p_tc = TYPECODE(obj);
  return TCL_OK;
}

#if WANT_TYPE_COMMAND
/*-----------------------------------------------------------------------
 *	Test command
 *-----------------------------------------------------------------------*/
static int TypeCode_cmd(ClientData clientData, Tcl_Interp *interp,
			int objc, Tcl_Obj *CONST objv[])
{
  char *opt;
  Tcl_Obj *obj;

  if (objc != 3) goto usage;
  opt = Tcl_GetStringFromObj(objv[1], NULL);
  obj = objv[2];

  if (strcmp(opt, "kind")==0) {
    MUSAGE(1,"typecode");
    if (obj->typePtr != &torb_TypeCode_objtype) {
      Tcl_AppendResult(interp, "not a TypeCode", (char*)0);
      return TCL_ERROR;
    }
    Tcl_SetObjResult(interp, Tcl_NewIntObj(TYPECODE(obj)->kind));
      
  } else if (strcmp(opt, "force")==0) {
    MUSAGE(1,"typecode");
    if (Tcl_ConvertToType(interp, obj, &torb_TypeCode_objtype) != TCL_OK) {
      return TCL_ERROR;
    }
    Tcl_IncrRefCount(obj);
    Tcl_SetObjResult(interp, obj);

  } else if (strcmp(opt, "setx")==0) {
    if (Tcl_ConvertToType(interp, obj, &torb_TypeCode_objtype) != TCL_OK) {
      return TCL_ERROR;
    }
    Tcl_IncrRefCount(obj);
    Tcl_ObjSetVar2(interp, Tcl_NewStringObj("x", -1), NULL,
		   obj, 0);

  } else if (strcmp(opt, "name")==0) {
    MUSAGE(1,"typecode");
    if (Tcl_ConvertToType(interp, obj, &torb_TypeCode_objtype) != TCL_OK) {
      return TCL_ERROR;
    }
    Tcl_SetResult(interp, TYPECODE(obj)->name, TCL_VOLATILE);

  } else if (strcmp(opt, "id")==0) {
    MUSAGE(1,"typecode");
    if (Tcl_ConvertToType(interp, obj, &torb_TypeCode_objtype) != TCL_OK) {
      return TCL_ERROR;
    }
    Tcl_SetResult(interp, TYPECODE(obj)->repo_id, TCL_VOLATILE);

  } else if (strcmp(opt, "regen")==0) {
    MUSAGE(1,"typecode");
    if (Tcl_ConvertToType(interp, obj, &torb_TypeCode_objtype) != TCL_OK) {
      return TCL_ERROR;
    }
    Tcl_InvalidateStringRep(obj);
    Tcl_SetObjResult(interp, obj);

  } else {
    goto usage;
  }
  return TCL_OK;

 usage:
  Tcl_AppendResult(interp, "usage: ", Tcl_GetStringFromObj(objv[0], NULL),
		   " kind|force|name|id obj", (char*)0);
  return TCL_ERROR;
}

/*-----------------------------------------------------------------------
 *	Initialisation
 *	Add the TypeCode object to the interpreter
 *-----------------------------------------------------------------------*/
int Typecode_Init(Tcl_Interp *interp)
{
  Tcl_RegisterObjType(&torb_TypeCode_objtype);
  Tcl_CreateObjCommand(interp, "type", TypeCode_cmd, NULL, NULL);
  return TCL_OK;
}
#endif
